use serde::{Deserialize, Serialize};
use serde_json::Value;
use std::collections::{HashMap, HashSet};

use super::{Commands, Operation};

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct Workflow {
    pub name: String,
    pub inputs: Vec<String>,

    #[serde(rename = "step")]
    steps: Vec<WorkflowStep>,
}

pub type StringMap = std::collections::HashMap<String, String>;

#[derive(Serialize, Deserialize, Clone, Debug)]
pub struct WorkflowStep {
    r#type: String,

    #[serde(default)]
    blob: StringMap,

    #[serde(default)]
    values: StringMap,

    #[serde(default)]
    inputs: StringMap,

    #[serde(default)]
    outputs: StringMap,
}

#[derive(Clone, Debug)]
struct InvocableOperation {
    name: String,
    binary: String,
    operation: Operation,
}

impl Workflow {
    /// Generate a [`clap::Command`] for a [`Workflow`], where the inputs can be defined either by
    /// command-line arguments or via a JSON input file.
    pub fn generate_command(&self) -> clap::Command {
        let mut command = clap::Command::new(&self.name).arg(clap::arg!(
                --"input-file" [FILE]
                "A file containing any inputs not passed on the command line"
        ));
        for input in &self.inputs {
            let arg = clap::Arg::new(input)
                .required(false)
                .long(input.replace('_', "-"))
                .value_name(input.to_uppercase());
            command = command.arg(arg);
        }
        command
    }

    fn load_inputs(&self, matches: &clap::ArgMatches) -> StringMap {
        let mut map = StringMap::default();
        let input_file: Option<StringMap> = matches
            .get_one::<std::path::PathBuf>("input-file")
            .and_then(|p| std::fs::File::open(p).ok())
            .and_then(|f| serde_json::from_reader(f).ok());
        for input in &self.inputs {
            match matches.get_one::<String>(input) {
                Some(value) => {
                    map.insert(input.clone(), value.clone());
                    continue;
                }
                None => {
                    if let Some(value) = input_file.as_ref().and_then(|f| f.get(input)) {
                        map.insert(input.clone(), value.clone());
                        continue;
                    }
                }
            }
            panic!("Key was not found: {input}");
        }

        map
    }

    fn simulate_workflow(&self, mut data: HashSet<String>, operations: &[InvocableOperation]) {
        // simulate the steps by using a HashSet to traverse the inputs and outputs and ensure
        // there's no inconsistencies
        for (i, step) in self.steps.iter().enumerate() {
            // NOTE: overflow possible but unlikely
            let step_index = i + 1;
            let step_type = &step.r#type;

            // Find the relevant Operation
            let Some(invocable) = operations.iter().find(|op| op.name == *step_type) else {
                panic!("Could not find operation: {step_type}");
            };

            // Check if we have the keys we want to pass into the module.
            for in_memory_name in step.inputs.values() {
                if !data.contains(in_memory_name) && !step.values.contains_key(in_memory_name) {
                    panic!("Failed simulation: step #{step_index} ({step_type}): missing value {in_memory_name}");
                }
            }

            // Check that the module accepts those keys.
            for module_input_name in step.inputs.keys() {
                if !invocable
                    .operation
                    .arguments
                    .iter()
                    .any(|arg| *module_input_name == arg.name)
                {
                    eprintln!("Simulation: step #{step_index} ({step_type}): input value {module_input_name} will be passed through as JSON input");
                }
            }

            // Add the keys we get from the module.
            for in_memory_name in step.outputs.values() {
                data.insert(in_memory_name.clone());
            }
        }
    }

    pub fn handle(&self, matches: &clap::ArgMatches, modules: Commands) {
        let inputs = self.load_inputs(matches);
        let data: HashMap<String, Value> = inputs
            .into_iter()
            .map(|(k, v)| (k, Value::String(v)))
            .collect();

        let mut operations = vec![];

        for (module_name, module_binary, module_operations) in modules {
            for operation in module_operations {
                let operation_name = &operation.name;
                let io = InvocableOperation {
                    name: format!("{module_name}-{operation_name}"),
                    binary: module_binary.clone(),
                    operation: operation.clone(),
                };
                operations.push(io);
            }
        }

        if matches.get_flag("simulate-workflow") {
            self.simulate_workflow(data.into_keys().collect(), &operations);
            return;
        }

        todo!("Unsimulated transaction!");
    }
}
